from typing import *
import re

from parsee.extraction.extractor_elements import ExtractedEl, FinalOutputTableColumn
from parsee.extraction.tasks.meta_info_structuring.meta_info import MetaInfoModel
from parsee.extraction.models.llm_models.structuring_schema import get_prompt_schema_item
from parsee.templates.general_structuring_schema import StructuringItemSchema
from parsee.extraction.extractor_dataclasses import ParseeMeta
from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel
from parsee.storage.interfaces import StorageManager
from parsee.extraction.tasks.meta_info_structuring.features import LLMMetaFeatureBuilder


class MetaLLMModel(MetaInfoModel):

    def __init__(self, items: List[StructuringItemSchema], llm: LLMBaseModel, storage: StorageManager, **kwargs):
        super().__init__(items)
        self.storage = storage
        self.model_name = llm.spec.internal_name if llm is not None else "llm"
        self.default_prob_answer = 0.8
        self.elements = []
        self.llm = llm
        self.feature_builder: LLMMetaFeatureBuilder = LLMMetaFeatureBuilder()

    def parse_prompt_answer(self, prompt_answer: str) -> Dict[str, Tuple[str, bool]]:
        answers = prompt_answer.splitlines()
        output = {}
        for k, answer in enumerate(answers):
            result = re.search(r'((\d+)\) *)(.+)', answer)
            if result is not None and len(result.groups()) > 2:
                number = result.group(2)
                value_predicted = result.group(3)
                if number.isdigit() and 0 <= int(number)-1 < len(self.items):
                    item_idx = int(number)-1
                    output[self.items[item_idx].id] = get_prompt_schema_item(self.items[item_idx]).get_value(value_predicted)
        return output

    def predict_meta(self, columns: List[FinalOutputTableColumn], elements: List[ExtractedEl]) -> List[List[ParseeMeta]]:

        all_output = []
        for column in columns:
            prompt = self.feature_builder.make_prompt(column, elements, self.items)

            prompt_answer, amount = self.llm.make_prompt_request(prompt)

            self.storage.log_expense(self.llm.spec.internal_name, amount, "meta LLM")

            prediction_dict = self.parse_prompt_answer(prompt_answer)

            output: List[ParseeMeta] = []
            for key, values in prediction_dict.items():
                value, parse_success = values
                output.append(ParseeMeta(self.model_name, column.col_idx, column.sources, key, value, self.default_prob_answer if parse_success else 0))
            all_output.append(output)

        return all_output



from typing import *

from parsee.converters.image_creation import ImageCreator
from parsee.cloud.api import ParseeCloud
from parsee.extraction.extractor_elements import StandardDocumentFormat, ExtractedEl
from parsee.extraction.extractor_dataclasses import Base64Image


class CloudImageFetcher(ImageCreator):

    def __init__(self, cloud: ParseeCloud):
        self.cloud = cloud

    def get_images(self, document: StandardDocumentFormat, element_selection: List[ExtractedEl], max_images: Optional[int], max_image_size: Optional[int]) -> List[Base64Image]:

        # collect all relevant pages
        page_indexes = []
        for el in element_selection:
            if el.source.other_info is not None and "page_idx" in el.source.other_info and int(el.source.other_info["page_idx"]) not in page_indexes and (
                    max_images is None or len(page_indexes) < max_images):
                page_indexes.append(int(el.source.other_info["page_idx"]))

        return [self.cloud.get_image(document.source_identifier, page_index, max_image_size) for page_index in page_indexes]



from typing import *
from decimal import Decimal

from parsee.extraction.extractor_dataclasses import ParseeBucket
from parsee.templates.element_schema import ElementSchema
from parsee.templates.mappings import MappingSchema
from parsee.datasets.dataset_dataclasses import MappingUniqueIdentifier
from parsee.extraction.tasks.mappings.features import MappingFeatureBuilder
from parsee.extraction.extractor_elements import FinalOutputTable


class MappingModel:

    model_name = ""
    feature_builder: MappingFeatureBuilder

    def __init__(self, items: List[ElementSchema], **kwargs):
        self.items = items
        self.memory: Dict[str, List[ParseeBucket]] = {}

    def classify_with_schema(self, table: FinalOutputTable, schema: MappingSchema) -> List[ParseeBucket]:
        raise NotImplemented

    def classify_elements(self, table: FinalOutputTable) -> Tuple[List[ParseeBucket], Union[None, MappingSchema]]:
        for item in self.items:
            if item.id == table.detected_class and item.mapRows is not None:
                mapping_schema = item.mapRows
                full_signature = f"{mapping_schema.id}_{table.li_identifier}"
                if full_signature in self.memory:
                    return self.memory[full_signature], mapping_schema
                else:
                    self.memory[full_signature] = self.classify_with_schema(table, mapping_schema)
                    return self.memory[full_signature], mapping_schema
        return [], None



from typing import *

from parsee.extraction.extractor_elements import StandardDocumentFormat, ExtractedEl


class VectorStore:

    def find_closest_elements(self, document: StandardDocumentFormat, search_element_title: str, keywords: str, tables_only: bool = True) -> List[ExtractedEl]:
        raise NotImplemented


from typing import List, Optional, Union, Tuple
import re
from decimal import Decimal

from parsee.templates.general_structuring_schema import StructuringItemSchema
from parsee.utils.enums import OutputType
from parsee.utils.helper import clean_numeric_value_llm, get_entity_value, get_date_regex, clean_numeric_value


NOT_AVAILABLE: str = "n/a"


class PromptSchemaItem:

    def __init__(self, possible_values: List[str], example: Optional[str] = None,
                 default_value: Optional[str] = None):
        self.possible_values = possible_values
        self.example = example
        self.default_value = default_value if default_value is not None and default_value.strip() != "" else None

    def is_valid_input(self, val: Union[str, None]) -> bool:
        return val is not None and val.strip() != ""

    def get_example(self, clean_value: bool = False) -> str:
        raise NotImplemented

    def get_possible_values_str(self) -> str:
        raise NotImplemented

    def get_default_value(self) -> str:
        return self.default_value if self.default_value is not None else NOT_AVAILABLE

    def get_value(self, value: str) -> Tuple[str, bool]:
        return value, True

    def parsed_to_raw(self, value: str) -> str:
        return value


class ListClassificationItem(PromptSchemaItem):

    def __init__(self, possible_values: List[str], example: Optional[str] = None,
                 default_value: Optional[str] = None):
        super().__init__(possible_values, example, default_value)

        self.possible_lowercase = [x.lower() for x in possible_values]

    def format_list_choice(self, val):
        return f"{val}"

    def get_example(self, clean_value: bool = False) -> str:
        return self.format_list_choice(self.possible_values[0]) if not clean_value else self.possible_values[0]

    def get_possible_values_str(self) -> str:
        return "possible values (select one from json array): [" + ", ".join(
            "'"+self.format_list_choice(x)+"'" for x in self.possible_values) + "]"

    def get_value(self, value: str) -> Tuple[str, bool]:

        value = value.lower()
        search = re.search(r'(\$( |)(.+)( |)\$)', value)
        if search is not None and len(search.groups()) == 4:
            value = search.groups()[2].strip()
        if value not in self.possible_lowercase:
            # try simple string matching
            for val in self.possible_lowercase:
                search = re.search(rf'\b{val}\b', value)
                if search is not None:
                    value = val
                    break
            if value not in self.possible_lowercase:
                return self.get_default_value(), False
        idx = self.possible_lowercase.index(value)
        return self.possible_values[idx], True

    def parsed_to_raw(self, value: str) -> str:
        return self.format_list_choice(value)


class PositiveIntegerItem(PromptSchemaItem):

    def __init__(self, example: Optional[str] = None, default_value: Optional[str] = None):
        super().__init__([], example, default_value)

    def get_example(self, clean_value: bool = False) -> str:
        return self.example if self.is_valid_input(self.example) else "123"

    def get_possible_values_str(self) -> str:
        return "possible values: positive integer"

    def get_value(self, value: str) -> Tuple[str, bool]:
        val = clean_numeric_value_llm(value)
        if val is None:
            return self.get_default_value(), False
        return str(abs(int(val))), True


class TextItem(PromptSchemaItem):

    def __init__(self, example: Optional[str] = None, default_value: Optional[str] = None):
        super().__init__([], example, default_value)

    def get_example(self, clean_value: bool = False) -> str:
        return self.example if self.is_valid_input(self.example) else "answer to question"

    def get_possible_values_str(self) -> str:
        return "possible values: any text"


class NumericItem(PromptSchemaItem):

    def __init__(self, example: Optional[str] = None, default_value: Optional[str] = None):
        super().__init__([], example, default_value)

    def get_example(self, clean_value: bool = False) -> str:
        return self.example if self.is_valid_input(self.example) else "123"

    def get_possible_values_str(self) -> str:
        return "possible values: any number (positive or negative, including decimals)"

    def get_value(self, value: str) -> Tuple[str, bool]:
        val = clean_numeric_value_llm(value)
        if val is None:
            return self.get_default_value(), False
        return str(val), True


class PercentageItem(PromptSchemaItem):

    def __init__(self, example: Optional[str] = None, default_value: Optional[str] = None):
        super().__init__([], example, default_value)

    def get_example(self, clean_value: bool = False) -> str:
        example = clean_numeric_value(self.example) if self.is_valid_input(self.example) else Decimal(0.15)
        if example < 1:
            example = example*100
        return str(round(float(example)))+"%"

    def get_possible_values_str(self) -> str:
        return "possible values: any percentage value"

    def get_value(self, value: str) -> Tuple[str, bool]:
        val = clean_numeric_value_llm(value)
        if val is None:
            return self.get_default_value(), False
        # check if value has to be multiplied
        mult = 1 if "%" in value or val > 1 else 100
        val = val * mult
        if val.is_integer():
            val = str(int(val))
        else:
            val = str(val)
        return val+"%", True


class EntityItem(PromptSchemaItem):

    def __init__(self, example: Optional[str] = None, default_value: Optional[str] = None):
        super().__init__([], example, default_value)

    def get_example(self, clean_value: bool = False) -> str:
        return self.example if self.is_valid_input(self.example) else "answer to question"

    def get_possible_values_str(self) -> str:
        return "possible values: any text, but please only provide the 'entity' (e.g. a thing, company or person, depending on the question) and nothing else"

    def get_value(self, value: str) -> Tuple[str, bool]:
        val = get_entity_value(value)
        if val is None:
            return value.strip().rstrip(".").strip(), True
        return val.strip().rstrip(".").strip(), True


class DateItem(PromptSchemaItem):

    def __init__(self, example: Optional[str] = None, default_value: Optional[str] = None):
        super().__init__([], example, default_value)

    def get_example(self, clean_value: bool = False) -> str:
        return self.example if self.is_valid_input(self.example) else "2023-11-24"

    def get_possible_values_str(self) -> str:
        return "possible values: date in the format: YYYY-MM-DD"

    def get_value(self, value: str) -> Tuple[str, bool]:
        val = get_date_regex(value)
        if val is None:
            return self.get_default_value(), False
        return str(val), True


def get_prompt_schema_item(item: StructuringItemSchema) -> PromptSchemaItem:
    if item.type == OutputType.LIST:
        return ListClassificationItem(item.valuesList, item.example, item.defaultValue)
    elif item.type == OutputType.INTEGER:
        return PositiveIntegerItem(item.example, item.defaultValue)
    elif item.type == OutputType.NUMERIC:
        return NumericItem(item.example, item.defaultValue)
    elif item.type == OutputType.TEXT:
        return TextItem(item.example, item.defaultValue)
    elif item.type == OutputType.ENTITY:
        return EntityItem(item.example, item.defaultValue)
    elif item.type == OutputType.DATE:
        return DateItem(item.example, item.defaultValue)
    elif item.type == OutputType.PERCENTAGE:
        return PercentageItem(item.example, item.defaultValue)
    raise Exception("item not found")



# This is also defined in setup.py and must be updated in both places.
__version__ = "0.1.0"

# top level imports
from parsee.cloud.api import ParseeCloud
from parsee.templates.helpers import MetaItem, StructuringItem, TableItem, create_template
from parsee.extraction.models.helpers import *
from parsee.converters.main import load_document, from_text
from parsee.extraction.run import run_job_with_single_model
from parsee.utils.enums import OutputType, DocumentType



from typing import List, Dict, Optional

from parsee.extraction.extractor_elements import ExtractedEl, FinalOutputTableColumn
from parsee.templates.general_structuring_schema import StructuringItemSchema
from parsee.extraction.extractor_dataclasses import ParseeMeta, ExtractedSource
from parsee.extraction.tasks.meta_info_structuring.features import MetaFeatureBuilder


class MetaInfoModel:

    model_name = ""
    feature_builder: MetaFeatureBuilder

    def __init__(self, items: List[StructuringItemSchema]):
        self.items = items

    def predict_meta(self, columns: List[FinalOutputTableColumn], elements: List[ExtractedEl]) -> List[List[ParseeMeta]]:
        # returns meta information
        raise NotImplemented



import os
import time
from typing import List, Tuple
from dataclasses import dataclass
from decimal import Decimal
import math

import tiktoken
import ollama
from ollama import Client

from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel, get_tokens_encoded, truncate_prompt
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.extraction.models.llm_models.prompts import Prompt


class OllamaModel(LLMBaseModel):

    def __init__(self, model: MlModelSpecification):
        super().__init__(model)
        self.max_retries = 0
        self.encoding = tiktoken.get_encoding("cl100k_base")
        self.max_tokens_answer = 1024
        self.max_tokens_question = self.spec.max_tokens - self.max_tokens_answer

        self.client = Client(host='http://localhost:11434' if model.file_path is None else model.file_path)

    def _call_api(self, prompt: str, retries: int = 0, wait: int = 5) -> str:
        try:
            response = self.client.chat(model=self.spec.internal_name, messages=[
                {
                    'role': 'user',
                    'content': prompt,
                },
            ])
            return response["message"]["content"]
        except Exception as e:
            if retries < self.max_retries:
                time.sleep(wait * 2 ** retries)
                return self._call_api(prompt, retries + 1)
            else:
                return ""

    def make_prompt_request(self, prompt: Prompt) -> Tuple[str, Decimal]:
        final_prompt, num_tokens_input = truncate_prompt(str(prompt), self.encoding, self.max_tokens_question)
        return self._call_api(final_prompt), Decimal(0)
